/**
 * A class providing basic colordinate functionality
 * @author David M. Smith
 */
 
import java.awt.*;

public class Coord {
  public double x;
  public double y;

  /**
   * Construct a new coordinate
   * @param p a Point
   */
  public Coord( Point p ) {
    x = p.x;
    y = p.y;
  }
	
  /**
   * Construct a new coordinate
   * @param px an integer
   * @param py an integer
   */
  public Coord( double dx, double dy ) {
    x = dx;
    y = dy;
  }
	
  /**
   * Construct a new coordinate
   * @param px an integer
   * @param py an integer
   */
  public Coord( int px, int py ) {
    x = px;
    y = py;
  }
	
  /**
   * project the difference between the back of a queue and
   * the location of this vehicle onto the direction dir.
   * @param where a Coord - the vehicle location
   * @param dir a Coord - the direction of the street
   * @return the projection
   */
  public static double relPos( Coord back, Coord where, Coord dir ) {
    double res = dir.scalarProd( where.diff(back) ); 
    //		if(Global.DEBUG)
    //			Global.debug.println("relPos(" + back + "," + where 
    //						+ "," + dir + ") -> " + res );
    return res;
  }

  /**
   * compute the scalar product of this Coord and another
   * @param a Coord - the other vector
   * @return double the scalar product
   */
  public double scalarProd(Coord b) {
    return (x * b.x) + (y * b.y);
  }

  /**
   * Construct a new coordinate normalizing this one
   * @return normalized coordinate
   */
  public Coord normalize() {
    double mag = Math.sqrt( x*x + y*y );
    return new Coord( x/mag, y/mag );
  }
	
  /**
   * Construct a new coordinate normal to this Coord
   * @return resulting normal
   */
  public Coord normal() {
    return new Coord( y, -x );
  }
	
  /**
   * Construct a new coordinate summing two Coords
   * @param c the addend
   * @return resulting coordinate
   */
  public Coord sum(Coord c) {
    return new Coord( x + c.x, y + c.y );
  }
	
  /**
   * Construct a new coordinate differencing two Coords
   * @param c the subtrahend
   * @return resulting coordinate
   */
  public Coord diff(Coord c) {
    return new Coord( x - c.x, y - c.y );
  }
	
  /**
   * Construct a new coordinate reversing this one
   * @return reversed coordinate
   */
  public Coord negate() {
    return new Coord( -x, -y );
  }
	
  /**
   * Construct a new coordinate scaling this one to the map
   * @return scaled coordinate
   */
  public Coord screenScale() {
    return new Coord( (x * Global.SCALE) + Global.X_OFFSET, 
		      (y * Global.SCALE) + Global.Y_OFFSET );
  }
	
  /**
   * Construct a new coordinate scaling this one from the map
   * @return unscaled coordinate
   */
  public Coord unScale() {
    return new Coord( (x - Global.X_OFFSET) / Global.SCALE, 
		      (y - Global.Y_OFFSET) / Global.SCALE );
  }
	

  /**
   * Construct a new coordinate scaling this one by a constant
   * @param sc constant scale
   * @return scaled coordinate
   */
  public Coord scale(double sc) {
    return new Coord( x * sc, 
		      y * sc );
  }
	

  /**
   * Construct a new coordinate perpendicular to this one
   * @param c the other location
   * @param s separation between the coords
   * @return perpendicular
   */
  public Coord perp(Coord c, double s) {
    return new Coord( (y - c.y) / s, 
		      (c.x - x) / s );
  }
	

  /**
   * Compute the distance from this coordinate to another
   * @param c the other coordinate
   * @return distance between them
   */
  public double distance(Coord c) {
    double dx = x - c.x;
    double dy = y - c.y;

    return Math.sqrt((dx*dx) + (dy*dy));
  }
	
  /**
   * Compute the angle from this coordinate to another
   * @param c the other coordinate
   * @return distance between them
   */
  public double angleTo(Coord c) {
    double dx = c.x - x;
    double dy = c.y - y;

    return Math.atan2(dy, dx);
  }
	

  /**
   * document the contents of the coordinate
   * @return the string describing the coordinate
   */
  public String toString() {
    return "[ " + x + ", " + y + " ]";
  }

}


