/**
 * A class providing basic street functionality
 * @author David M. Smith
 */
 
import java.awt.*;
import java.lang.*;

public class Street extends TrafficObject {

  protected String name;
  protected int limit;
  protected double width;
  protected Color color;


  public Coord[]  location = { null, null };
  public Coord[] direction = { null, null };
  protected Coord[] qBack = { null, null };
  protected TravQueue[] driving = { null, null };
  private TravQueue[] waiting = { null, null };

  /**
   * Construct a new street 
   */
  public Street(String nm, int lim ) {
    int i;

    name = nm;
    limit = lim;
    width = Global.STREET_WIDTH;
    color = Color.green;
    if( limit > 50 ) {
      width = width * 2;
      color = Color.red;
    }


    for( i = 0; i < 2; i++ ) {
      driving[i] = new TravQueue();
      waiting[i] = new TravQueue();
    }
    //		if(Global.DEBUG) { 
    //		 	Global.debug.println("Created " + toString() );
    //		}
  }
	
  /**
   * draw the street
   * @param g the graphics environment
   */
  public void draw(Graphics g) {
    Coord from = location[0].screenScale();
    Coord   to = location[1].screenScale();
    double size = from.distance(to);
    Coord unit_normal = from.perp( to, size );
    int i;

    Coord leftFrom  =  from.sum( unit_normal.scale(width) ); 
    Coord leftTo    =    to.sum( unit_normal.scale(width) ); 
    Coord rightFrom = from.diff( unit_normal.scale(width) ); 
    Coord rightTo   =   to.diff( unit_normal.scale(width) ); 

    g.setColor(color);
    g.drawLine( (int)  leftFrom.x, (int)  leftFrom.y, 
		(int)  leftTo.x, (int)  leftTo.y );
    g.drawLine( (int) rightFrom.x, (int) rightFrom.y, 
		(int) rightTo.x, (int) rightTo.y );
    for( i = 0; i < 2; i++ ) {
      driving[i].draw(g);
      waiting[i].draw(g);
    }
  }

  /**
   * determine if a street driving list is full
   * @param d the direction (0 or 1)
   * @return true or false
   */
  public boolean isFull( int d, Vehicle v) {
    boolean answer = ( Coord.relPos( qBack[d], location[1-d], direction[d] ) > -1.0 );

    //	if( Global.DEBUG ) 
    //	Global.debug.println( "Street " + toString() + "[" + d + "] full? -> " + answer );
    return answer;
  }

  
  /**
   * determine if a street waiting list is empty
   * @param d the direction (0 or 1)
   * @return true or false
   */
  public boolean isEmpty( int d ) {
    boolean res = waiting[d].isEmpty();
    Vehicle v = null;

    if( !res ) {
      try {
	v = (Vehicle) waiting[d].peek();
      } catch (Exception e) {
	System.out.println("Exception: " + e.getMessage() );
	Global.debug.println("Exception: " + e.getMessage());
      }
      //  if(Global.DEBUG) {
      //    Global.debug.println("Vehicle front check");
      //    Global.debug.println(" - street loc[" + d + "] is " + location[d]);
      //    Global.debug.println(" - vehicle loc is " + v.where );
      //    Global.debug.println(" - vehicle distance is " 
      //                         + v.where.distance(location[d]) );
      // }
      if(limit > 50) {
	res = (v.where.distance(location[d]) > (2*Global.AT_FRONT) );
      } else {
	res = (v.where.distance(location[d]) > Global.AT_FRONT );
      }
    }
    return res;
  }


  /**
   * dequeue from the street waiting list
   * @param d the direction (0 or 1)
   * @param o the object to enqueue
   */
  public Vehicle dequeue( int d ) {
    return (Vehicle) waiting[d].dequeue();
  }


  /**
   * peek into the street waiting list
   * @param d the direction (0 or 1)
   * @param o the object to look at
   */
  public Vehicle peek( int d ) {
    return (Vehicle) waiting[d].peek();
  }


  /**
   * enqueue onto the street driving list
   * @param o the object to enqueue
   * @param d the direction (0 or 1)
   */
  public void enqueue( Vehicle v, int d) {
    Coord delta = direction[d].normal().scale( width/(2*Global.SCALE) );

    driving[d].enqueue(v);
    v.locate( location[1-d].diff(delta), direction[d] );
    v.update( v.getLength() + Global.STREET_WIDTH/Global.SCALE );
  }


  public void clear() {
		for(int i = 0; i < 2; i++) {
			driving[i].clearQueue();
			waiting[i].clearQueue();
		}
  }
  /**
   * update the street
   * @param dt the time difference in seconds
   */
  public void update(double dt) {
    int i;

    for(i = 0; i < 2; i++ ) {// move all the vehicles
      waiting[i].update(dt * limit / 2);
      driving[i].update(dt * limit);
    }
    for(i = 0; i < 2; i++ ) {
      if(Global.DEBUG)
	Global.debug.println("Street.update " + i);
      try {
	qBack[i] = separate( waiting[i],
			     driving[i],
			     direction[1-i],
			     location[i].sum( direction[1-i].normal().scale( width/(2*Global.SCALE) )));
      } catch( Exception e) { 
	System.out.print("Error " + e + " updating a Street" ); 
	Global.debug.println("Error " + e + " updating a Street" ); 
      } 
    }
  }


  /**
   * separate the vehicles on the two queues
   * @param waiting a TravQueue stationary at the end of the segment
   * @param moving a TravQueue moving at the speed limit
   * @param dir a Coord, the direction of motion
   * @param hd a Coord, the current head of the queues 
   * @return a Coord, the latest back of the queue
   */
  public Coord separate( TravQueue stopped, TravQueue moving, 
			 Coord dir, Coord hd )  {
    //		if(Global.DEBUG)
    //		  Global.debug.println(toString() + ".separate( st, mv, " + dir + ", " + hd + " )" );
    Coord back = stopped.separate( 0, dir, hd.sum(dir.scale(2*width/Global.SCALE)) );
    Coord res  = moving.separate( limit, dir, back );
    Vehicle v;
 
    if( !moving.isEmpty() ) {
      v = (Vehicle) moving.peek();
      while (!moving.isEmpty() && v.stopped ) {
	//				if(Global.DEBUG)
	//					Global.debug.print("Separate moved " + moving.peek() 
	//							+ " from " + moving );
	stopped.enqueue( moving.dequeue() );
	//				if(Global.DEBUG)
	//					Global.debug.println(" to  " + stopped );
	if( !moving.isEmpty() )
	  v = (Vehicle) moving.peek();
      } ;
    }
    return res;
  }


  /**
   * define the location of one end of the street
   * @param pt the location
   * @param index select which end (0 or 1)
   */
  public void locate(Coord pt, int index) {
    location[index] = pt;
    //		if(Global.DEBUG) {
    //		 	Global.debug.println(toString() + "[" + index + "] located at " 
    //							+ location[index] );
    //		}
    if( location[1 - index] != null ) {
      direction[1] = location[1].diff(location[0]).normalize();
      direction[0] = direction[1].negate();
      //			if(Global.DEBUG) {
      //			 	Global.debug.println(toString() + " direction is " +
      //							direction[0] );
      //			}
    }
  }


  /**
   * document the contents of the street
   * @return the string describing the street
   */
  public void display() {
    int i;

    for( i = 0; i < 2; i++ ) {
      Global.debug.println(toString() + "[" + i + "]");
      Global.debug.println("location:" + location[i]
			   + "; direction:" + direction[i]
			   + "; qBack:" + qBack[i] );
      Global.debug.println("driving:" + driving[i] );
      Global.debug.println("waiting:" + waiting[i] );
    }
  }


  /**
   * document the contents of the street
   * @return the string describing the street
   */
  public String toString() {
    return "Street: " + name;
  }
}


