/**
 * A class providing vehicle functionality at different
 * smartness levels
 * Smart levels are:
 *   0 - random turns, no restrictions
 *   1 - avoid U-turns
 *   2 - greedy path to destination
 *   3 - A* search for a path
 *   4 - optimal search for a path
 * @author David M. Smith
 */
 
import java.awt.*;

public class Vehicle extends TrafficObject {
  public static final int MAX_SMART = 3;

  private static int smarts = 2;

  private String make;
  private String model;
  protected double length;
  protected Coord heading;
  protected int ID;
  public Coord where;
  public boolean stopped;
  protected Street nextStreet;
  protected TravQueue history;
  protected int nextDir;
  public int smartLevel;
  protected Coord destination;

  /**
   * the source of unique vehicle IDs
   * incremented each time a new vehicle is created
   */
  private static int VehicleID = 1;

  /**
   * to create a base vehicle class with at least this much smarts
   * @param mk make of the vehicle
   * @param md model of the vehicle
   * @param l length of the vefhicle
   * @param p location of the vehicle
   */
  public Vehicle(String mk, String md, double l, Point p, int minSmart) {
    Coord tl; // top left in world coords
    Coord br; // bottom right in world coords

    make = mk;
    model = md;
    length = Global.SIZE_SCALE * l / Global.SCALE;
    where = new Coord(p);
    heading = new Coord(0,0);
    stopped = false;
    ID = VehicleID++;
    smartLevel = minSmart; // + (int) ((smarts - minSmart) * Math.random());
    if( smartLevel < 2 ) {
      destination = null;
    } else {
      smarts = 2;
      tl = Global.TOP_LEFT.unScale();
      br = Global.screenSize.unScale();
      double nx = tl.x + Math.random() * (br.x - tl.x);
      double ny = tl.y + Math.random() * (br.y - tl.y);
      destination = new Coord( nx, ny );
      history = new TravQueue();
    }
  }

  /**
   * to open the window for smart vehicles
   */
  public static void bumpSmarts() {
    smarts = MAX_SMART;
  }

  /**
   * to create a base vehicle class
   * @param mk make of the vehicle
   * @param md model of the vehicle
   * @param l length of the vefhicle
   * @param p location of the vehicle
   */
  public Vehicle(String mk, String md, double l, Point p) {

    this(mk, md, l, p, 0);
  }


  protected static Coord RxRy = new Coord( 5, 5);
  /**
   * draw a vehicle
   * @param g the graphics environment
   * @param color the color to draw it
   */
  public void draw(Graphics g, Color color) {
    Coord back = where.diff( heading.scale(0.75*length) ).screenScale();
    Coord front = where.screenScale();

    g.setColor(color);
    g.drawLine( (int) front.x, (int) front.y, 
		(int) back.x, (int) back.y );
    if(getDestination() != null) {
      back = getDestination().screenScale();
      g.setColor(Color.yellow);
      g.drawLine( (int) front.x, (int) front.y, 
		  (int) back.x, (int) back.y );
      back = back.diff(RxRy);
      g.drawOval( (int) back.x, (int) back.y, 10, 10); 
    }
  }


  /**
   * turn a vehicle
   * @param here the intersection where the vehicle is located
   * @param strt the street it is currently on
   * @param dir the direction it is facing
   */
  public void turn(Intersection here, Street strt, int dir) {
    int choice;
    int which;

    switch( smartLevel ) {
    case 0: // not smart at all
      choice = (int) ( here.streetIndex.length * Math.random() );
      which = here.streetIndex[choice];
      nextStreet = (Street) Global.streets.find( which );
      nextDir = 1 - here.directionIndex[choice];
      break;
    case 1: // avoid U-turns
      if( here.streetIndex.length < 2 ) {
	nextStreet = null;
      } else {
	do {
	  choice = (int) ( here.streetIndex.length * Math.random() );
	  which = here.streetIndex[choice];
	  nextStreet = (Street) Global.streets.find( which );
	  nextDir = 1 - here.directionIndex[choice];
	} while (nextStreet == strt);
      }
      break;
    case 2: // Greedy
      double best = 1000;
      double angleToDest = where.angleTo(getDestination());
      Street trySt;
      int tryDir;
      Coord otherEnd;
      double angleToTry;
      double diff;
      boolean already = isInHistory(here);
		if(Global.DEBUG)
        Global.debug.println("Try " + here + " >> " + already );
      if( already ) {
	nextStreet = null;
	smarts = MAX_SMART;
      } else {
	for( choice = 0; choice < here.streetIndex.length; choice++ ) {
	  which = here.streetIndex[choice];
	  trySt = (Street) Global.streets.find(which);
	  tryDir = 1 - here.directionIndex[choice];
	  otherEnd = trySt.location[tryDir];
	  angleToTry = where.angleTo(otherEnd);
	  diff = Math.abs( angleToDest - angleToTry );
	  if( diff > Math.PI ) diff = 2*Math.PI - diff;
	  if( best > diff ) {
	    best = diff;
	    nextStreet = trySt;
	    nextDir = tryDir;
	  }
	  if( !nextStreet.isFull(nextDir, this) ) history.enqueue(here);
	}
      }
      break;
    default:
      throw new RuntimeException("Illegal smart level");
    }
  }

  public Coord getDestination() {
    return destination;
  }

  /**
   * isInHistory
   * determine if we have looped back to a previous destination
   * if not, add this intersection  to the history
   * @param current intersection
   * @return if this intersectiuon is in the history list
   */
  private boolean isInHistory( Object o ) {
    Object item = history.reset();
    while( item != null ) {
      if( item == o ) return true;
      item = history.getNext();
    }
    return false;
  }

  /**
   * access the next street
   * @return nextStreet - the street we are going into
   */
  public Street getNextStreet() {
    return nextStreet;
  }

  /**
   * access the next direction
   * @return nextDirection - the direction we are going into
   */
  public int getNextDirection() {
    return nextDir;
  }


  /**
   * update method moving the vehicle
   * @param distance the distance in map coordinates
   */
  public void update(double distance) {
    //		if(Global.DEBUG)
    //			Global.debug.println(toString() + ".update(" + distance + ")");
    locate( where.sum( heading.scale(distance) ), heading );
    stopped = false;
  }


  public double getLength(){
    return length;
  }

  public static int getTotal(){
    return VehicleID;
  }

  /**
   * dummy draw method to satisfy TrafficObject
   */
  public void draw(Graphics g) {
    throw new RuntimeException("Some vehicle child class doesn't have a draw(g) method");
  }

	
  /**
   * to set the state of a vehicle
   * @param p location of the vehicle
   * @param h heading of the vehicle
   */
  public void locate( Coord p, Coord h) {
    where = p;
    heading = h;
  }

	
  /**
   * shuffle this vehicle backwards
   * @param back a Coord locating the back of the car in front
   * @param speed a double showing the vehicle speed
   * @param dir a Coord giving the street direction
   * @return the position of the back of this car
   */
  public Coord shuffle(Coord back, double speed, Coord dir) {
    double delta = (speed / Global.SPEED_FACTOR);
    Coord newBack;

    if(Global.DEBUG) {
      Global.debug.print("shuffle " + toString() 
			 + " at " + where + "; back " + back );
    }
    if( Coord.relPos(back, where, dir) < delta ) {
      where = back.sum( dir.scale( delta ) );
      stopped = true;
      if( Global.DEBUG ) {
	Global.debug.print("; moved to " + where );
      }
    }
    newBack = where.sum( dir.scale(delta + length) );
    if( Global.DEBUG ) {
      Global.debug.println( "; back is now " + newBack );
    }
    return newBack;
  }

  private static boolean oj = false;
  private static boolean police1 = false;
  private static boolean pion = false;
  private static boolean limo = false;
  private static boolean sunday = false;
  private static boolean marta = false;
  private static boolean exp = false;
  private static boolean sb = false;
  private static boolean taxi = false;
  private static boolean jalopy = false;
  private static boolean fire = false;
  private static boolean schoolbus = false;
  private static boolean ambulance = false;
  private static boolean fbf = false;
  private static boolean fire2 = false;
  private static boolean emerg = false;
  private static boolean wide = false;
  private static boolean amb1 = false;
  private static boolean stall = false;
  private static boolean fedex = false;
  private static boolean bus = false;
  private static boolean parade = false;
  /**
   * to create a random vehicle
   * @param loc a Coord, its location
   * @param hdg a Coord, its direction 
   * @return a new vehicle
   */
  public static Vehicle randomVehicle() {
    double x = Math.random();
    int index = (int) (30 * x);
    double factor = Math.random();
    Vehicle res = null;

    if( index < 1 ) {
      // make a cycle
      res = new Cycle( "Triumph", "1954", (int)(8 + factor*4), 0, 0 ); 
    } else if ( (index < 4) /*&& !madeOne*/ ) {
      // make a semi
      res = new Semi( "International", "2000", 
		      (int) (20 + factor*10), 
		      80, 0, 0, 0 ); 
    } else if ( index < 6 ) {
      // make a transporter
      res = new Transporter( "International", "1994", 
			     (int)(80 + factor*60), 
			     (int) (factor * 10), 0, 0 );
	 } else if ( index == 6) {
      switch( Global.chaseState ) {
      case 0: // not started
	       res = new GetawayCar( "Ford", "Pinto", 
			      (int) (20 + factor*10), 0, 0, 0 );
	       Global.chaseState++;
	       break;
      case 1: // add a police car
      case 2: // add a police car
	       res = new PoliceCar( "Jaguar", "XK8", 
			     (int) (20 + factor*10), 0, 0, 0 );
	       Global.chaseState++;
	       break;
      default:
	       res = new Car( "Lexus", "2000", 
		       (int) (20 + factor*10), 0, 0, 0 ); 
      } 
	 } else if (!police1 && index == 7) {
      res = new Policecar1( "Mustang", "2002", 
		     (int) (20 + factor*10), 0, 0, 0 ); 
			  police1 = true;
     } else if ( !oj && index == 8) {
     // make an OJ
     res = new OJ( "Bronco", "1994", (int) (25 + factor*10), 0, 0);
	  oj = true;
	 } else if ( !bus && index == 8) {
	  		res=new Bus("International","1976",60,(int)(20+factor*60),0,0);
			bus = true;
	 } else if ( !pion && 
	 					oj && 
						police1 && 
						sunday && 
						limo && 
						marta && 
						sb && 
						taxi && 
						jalopy && 
						fire && 
						schoolbus && 
						ambulance && 
						fbf && 
						fire2 && 
						emerg && 
						wide && 
						amb1 && 
						stall && 
						fedex && 
						parade && 
						(Global.chaseState > 2) &&
						bus &&
					   index == 9) {
	  		res=new PauloIon("International","1976",60,(int)(20+factor*60),0,0);
			pion = true;
    }else if( !limo && (index == 10) ){
	   limo = true;
    	res = new Limo( "Pimp Limo", "1969", (int)(20 + factor*20), 0, 0, 0) ;
    } else if (!sunday && (index == 11)) {
   // make a Sunday Driver
		  sunday = true;
        res = new Sunday("Cadillac", "2003", (int)(25 + factor*30), 0, 0);
    } else if( !marta && (index == 12) ) {
      // make a bus
      res = new Marta_Bus( "FlexBus", "1954", (int)(12 + factor*10), 0, 0 ); 
		marta = true;
    } else if( !exp && (index == 13) ) {
      // make an exploding car
       res = new ExplodingCar( "Fiat", "1988", (int)(20 + factor*10), 0, 0, 0 );
		exp = true;
	} else if (!sb && (index == 14) ) {
      res = new Schoolbus( "VW", "1975", 
		     (int) (80 + factor*60), 0, 0, 0 );
		sb = true;
	} else if (!taxi && (index == 15) ) {
      res = new Taxi( "VW", "1975", 
		     (int) (80 + factor*60), 0, 0 );
		taxi = true;
     } else if ( !jalopy && (index == 16) ) {
      //make a jalopy
      res = new Jalopy( "Geo", "Metro", 10, (int) (1 + factor*18), 0, 0);
		jalopy = true;
    } else if ( !fire && (index == 16) ) {
      // make a firetruck with ambulance
      res = new Firetruck( "FD-ATL", "2003", 
			     (int)(50 + factor*60), 
			     (int) (factor * 10), 0, 0 );
				  fire = true;
    }    else if ( !schoolbus && index == 17) {
     // make an schoolBus
     res = new SchoolBus1( "International", "2003", 100, 75,0, 0);
	  schoolbus = true;
     }else if ( !ambulance && (index == 18) ) {
	 //make Ambulance
	 res = new Ambulance("Pickup", "2003", (int) (25 + factor*10), 0, 0, 0);
	 ambulance = true;
    } else if ( !fbf && (index == 19) ) {                                /* Our vehicle class '4X4 Truck' */
     // make a 4x4 truck
      res = new FourByFourTruck( "Bronco", "1974", 
			 (int) (20 + factor*15), 0, 0, 0 );
 	   fbf = true;
    } else if (!fire2 && (index == 19 ) ) {
    // make a firetruck
      res = new Firetruck2("Hoser", "1989", (int) (70 + factor*50), 0, 0 );
		fire2 = true;
    } else if (!emerg && ( index == 20 ) ) {

// Added ambulance to random vehicle generator

      res = new Ambulance2( "Emergency", "Ambulance", 
			      (int) (20 + factor*10), 0, 0, 0 );
			      emerg = true;
    } else if ( !wide && (index == 21 ) ) {         // <<<<<<<<<<<< Wide Load Vehicle added
       // make a Wide Load
        res = new WideLoad("U-HAUL", "Mobile Home Mover",    // <<<<<< Wide Load information
                        (int) (25 + factor*30), 0, 0, 0);    // <<<<<< Wide Load Information
		  wide = true;
    } else if ( !amb1 && (index == 22 )) { /*&& !madeOne*/                //**New Code**(Ambulance)
      // make an ambulance
      res = new Ambulance1( "Patient Transport", "2003", 
		      (int) (20 + factor*10), 
		      80, 0, 0, 0 );  
      amb1 = true; // End of New Code**
    } else if ( !stall && (index == 23 ) ) {
     // make a StallCar
       res = new StallCar( "Impala", "1996", 
		     (int) (20 + factor*10), 0, 0, 0 ); 
			  stall = true;
    } else if (!fedex && (index == 24) ){
      res= new FedexTruck("Ford", "F-750", 
			(int)(30 + factor*14),0,0);
		fedex = true;
	  } else if ( !parade && index == 24) {
     // make an Parade
     res = new Parade1( "Parade", "1994", (int) (300 + factor*10), 200, 0, 0);
	  parade = true;
    } else {
      // make a car
      res = new Car( "Lexus", "2000", 
		     (int) (20 + factor*10), 0, 0, 0 ); 
    }
    if(Global.DEBUG) {
      Global.debug.println("Made a " + res);
      Global.debug.flush();
    }
    return res;
  }


  public String toString() {
    return "Vehicle ID: " + ID + ": " + make + ", " + model 
      + " length: " + length;
  }
}
